from typing import Dict, Tuple, Any, List, Optional

from drb import DrbNode, AbstractNode
from drb.exceptions import DrbNotImplementationException, DrbException
from drb.path import ParsedPath


class DrbXqueryItem(AbstractNode):

    def __init__(self, parent: DrbNode, name: str,
                 namespace_prefix: str, namespace_uri: str):
        super().__init__()
        self._attributes: Dict[Tuple[str, str], Any] = {}
        self._name = name
        self._namespace_uri = namespace_uri
        self._parent = parent
        self._children: List[DrbNode] = []
        self._path = None
        self._value = None
        self.order_elt = []
        self.prefix = namespace_prefix

    @property
    def parent(self) -> Optional[DrbNode]:
        return self._parent

    @property
    def path(self) -> ParsedPath:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def namespace_uri(self) -> Optional[str]:
        return self._namespace_uri

    @property
    def value(self) -> Optional[Any]:
        return self._value

    @property
    def attributes(self) -> Dict[Tuple[str, str], Any]:
        return self._attributes

    def get_attribute(self, name: str, namespace_uri: str = None) -> Any:
        key = (name, namespace_uri)
        if key in self.attributes.keys():
            return self.attributes[key]
        raise DrbException(f'Attribute not found name: {name}, '
                           f'namespace: {namespace_uri}')

    def add_attribute(self, name: str, value: any, namespace_uri: str = None):
        key = (name, namespace_uri)
        self.attributes[key] = value

    def set_attribute(self, attributes: dict):
        self._attributes = attributes

    def set_value(self, value: Any):
        self._value = value

    @property
    def children(self) -> List[DrbNode]:
        return self._children

    def has_impl(self, impl: type) -> bool:
        return False

    def get_impl(self, impl: type, **kwargs) -> Any:
        raise DrbNotImplementationException(f'no {impl} '
                                            f'implementation found')

    def append_child(self, node: DrbNode) -> None:
        self._children.append(node)

    def close(self):
        pass

    def get_named_child_list(self, name: str, namespace_uri: str = None) -> \
            List[DrbNode]:
        """
        Retrieves one or more children via its given name and its namespace.

        Parameters:
            name (str): child name
            namespace_uri (str): child namespace URI (default: None)
        Returns:
            List[DrbNode] - requested children
        Raises:
            TypeError: if item is not an int or a slice
            IndexError: if item is out of range of found children
            DrbException: if no child following given criteria is found
        """
        if self.namespace_aware or namespace_uri is not None:
            named_children = [x for x in self.children if x.name == name
                              and x.namespace_uri == namespace_uri]
        else:
            named_children = [x for x in self.children if x.name == name]
        if len(named_children) <= 0:
            raise DrbException(f'No child found having name: {name} and'
                               f' namespace: {namespace_uri}')
        return named_children

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.namespace_uri != other.namespace_uri:
            return False
        if self.value != other.value:
            return False

        return True

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)



import datetime
import calendar
import math
from decimal import Decimal

from drb import DrbNode

from .drb_xquery_context import DynamicContext
from .execptions import DynamicException, ErrorXQUERY


class DrbQueryFuncUtil:

    @staticmethod
    def convert_before_compare(left_item, right_item):
        left_value = DrbQueryFuncUtil.get_value(left_item)
        right_value = DrbQueryFuncUtil.get_value(right_item)

        if type(left_value) != type(right_value):
            if isinstance(left_value, (float, int, Decimal)):
                right_value = DrbQueryFuncUtil.get_numeric_value(
                    right_value)
            elif isinstance(right_value, (float, int, Decimal)):
                left_value = DrbQueryFuncUtil.get_numeric_value(
                    left_value)
        return left_value, right_value

    @staticmethod
    def is_equal(left_value, right_value):
        return left_value == right_value or \
               str(left_value) == str(right_value)

    @staticmethod
    def get_value(node, only_one=False):
        if isinstance(node, list):
            if len(node) == 0:
                return None
            if len(node) > 1 and only_one:
                raise DynamicException(ErrorXQUERY.FOAP0001,
                                       "func : requires one parameter")
            node = node[0]

        if isinstance(node, (DrbNode, DynamicContext)):
            return node.value

        return node

    @staticmethod
    def get_float_value(node):
        value = DrbQueryFuncUtil.get_value(node)
        try:
            if isinstance(value, datetime.datetime):
                value = DrbQueryFuncUtil.get_epoch(value)
            if isinstance(value, (float, Decimal, str, int)):
                return float(value)
        except Exception as error:
            pass

        raise DynamicException(ErrorXQUERY.XPTY0004,
                               "Error type " + value +
                               " can  not be converted to float")

    @staticmethod
    def get_int_value(node):
        value = DrbQueryFuncUtil.get_value(node)
        if value is None:
            return ''
        if isinstance(value, datetime.datetime):
            value = DrbQueryFuncUtil.get_epoch(value)

        if isinstance(value, int):
            return value
        try:
            if isinstance(value, (float, Decimal, str)):
                return int(value)
        except Exception as error:
            pass
        raise DynamicException(ErrorXQUERY.XPTY0004,
                               "Error type " + value +
                               " can  not be converted to integer")

    @staticmethod
    def get_decimal_value(node):
        value = DrbQueryFuncUtil.get_value(node)
        if value is None:
            return ''
        if isinstance(value, Decimal):
            return value
        try:
            if isinstance(value, datetime.datetime):
                value = DrbQueryFuncUtil.get_epoch(value)

            if isinstance(value, (int, str)):
                return Decimal(value)
            if isinstance(value, float):
                return Decimal.from_float(value)

        except Exception as error:
            pass
        raise DynamicException(ErrorXQUERY.XPTY0004,
                               "Error type " + value +
                               " can  not be converted to decimal")

    @staticmethod
    def get_byte_value(node):
        value = DrbQueryFuncUtil.get_value(node)
        try:
            if isinstance(value, (float, Decimal, str)):
                value = int(value)
            if isinstance(value, int):
                return value.to_bytes(1, byteorder='big')

        except Exception as error:
            raise DynamicException(ErrorXQUERY.XPTY0004,
                                   "Error type " + value +
                                   " can  not be converted to byte")

    @staticmethod
    def get_numeric_value(node, only_one=False):
        value = DrbQueryFuncUtil.get_value(node, only_one)

        if isinstance(value,  (float, Decimal, int)):
            return value
        if isinstance(value, str):
            value_to_test = value
            if value_to_test.startswith('-') or value_to_test.startswith('+'):
                value_to_test = value_to_test[1:]
            if value_to_test.isnumeric():
                return int(value)
            try:
                return float(value)
            except Exception as error:
                raise DynamicException(ErrorXQUERY.XPTY0004,
                                       "Error type " + value +
                                       " can  not be converted to numeric")

    @staticmethod
    def get_epoch(date_time):
        # return date_time.timestamp()
        return calendar.timegm(date_time.timetuple())

    @staticmethod
    def get_round_int_value(node):
        value = DrbQueryFuncUtil.get_value(node)
        if isinstance(value, int):
            return value
        if isinstance(value, (float, Decimal, str)):
            return round(float(value))

    @staticmethod
    def get_nane(node):
        if isinstance(node, list):
            if len(node) == 0:
                return ''
            node = node[0]
        if isinstance(node, DrbNode):
            return node.name
        elif isinstance(node, DynamicContext):
            return node.name

        return ''

    @staticmethod
    def get_node(node):
        if isinstance(node, list):
            if len(node) == 0:
                return ''
            node = node[0]

        if isinstance(node, (DrbNode, str)):
            return node
        elif isinstance(node, DynamicContext):
            return node.node

        return node

    @staticmethod
    def get_effective_boolean_value(node):
        if node is None:
            return False
        if isinstance(node, list):
            if len(node) == 0:
                return False
            if len(node) == 1:
                node = node[0]
            else:
                return True

        if isinstance(node, DrbNode):
            return True

        if isinstance(node, DynamicContext) \
                and (node.is_node() or node.is_attribute()):
            return True

        if isinstance(node, bool):
            return node

        if isinstance(node, str):
            return len(node) != 0

        if isinstance(node, (int, float, Decimal)):
            if node == 0:
                return False
            if isinstance(node, float) and math.isnan(node):
                return False
            return True
        raise DynamicException(ErrorXQUERY.FORG0006,
                               "Error result of is not a boolean" +
                               str(node))

    @staticmethod
    def get_boolean_value(node):
        node = DrbQueryFuncUtil.get_node(node)
        if node is None:
            return False
        if isinstance(node, DrbNode):
            return True
        value = DrbQueryFuncUtil.get_value(node)

        if isinstance(value, bool):
            return node
        if isinstance(value, str):
            if len(value) == 0:  # or value == '0':
                return False
            else:
                return True

        if isinstance(value, (int, float, Decimal)):
            if isinstance(value, float):
                if math.isnan(value):
                    return False
            if isinstance(value, Decimal):
                if value.is_nan():
                    return False

            if value != 0:
                return True
        return False

    @staticmethod
    def get_name_in_result(result) -> str:
        if result is None:
            raise DynamicException(ErrorXQUERY.XPTY0004,
                                   "Error exp for name "
                                   "return None")

        if isinstance(result, list):
            if len(result) == 1:
                result = result[0]
            else:
                raise DynamicException(ErrorXQUERY.XPTY0004,
                                       "Error exp for name "
                                       "return more than "
                                       "one element")
        if isinstance(result, (DrbNode, DynamicContext)):
            result = result.value
        try:
            return str(result)
        except Exception as error:
            raise DynamicException(ErrorXQUERY.XPTY0004,
                                   "Error exp for name "
                                   "return non string "
                                   "element")

    @staticmethod
    def get_string(node, concat_child=False) -> str:
        value_string = ''
        if isinstance(node, list):
            if len(node) == 0:
                return ''
            result = DrbQueryFuncUtil.get_string(node[0])
            for item in node[1:]:
                result = result + ' ' + DrbQueryFuncUtil.get_string(item)
            return result

        if isinstance(node, DynamicContext) and isinstance(node.node, DrbNode):
            node = node.node

        if isinstance(node, DynamicContext):
            value_string = node.value
        elif isinstance(node, DrbNode):
            value_string = node.value
            if value_string is None:
                value_string = ''
            if concat_child:
                for child in node.children:
                    value_string = value_string + \
                                   DrbQueryFuncUtil.get_string(child)
        elif isinstance(node, str):
            value_string = node
        elif isinstance(node, datetime.datetime):
            # value_string = node.strftime('%d-%b-%Y %H:%M:%S.%f')
            if node.microsecond != 0:
                value_string = node.strftime('%Y-%m-%dT%H:%M:%SZ%f')
            else:
                value_string = node.strftime('%Y-%m-%dT%H:%M:%SZ')
        elif isinstance(node, float):
            if node.is_integer():
                value_string = str(int(node))
            else:
                value_string = str(node)
        else:
            value_string = str(node)
        if value_string is None:
            return ''

        return str(value_string)

    @staticmethod
    def is_object_in_list(object_to_find, list_object: list):
        for item in list_object:
            if DrbQueryFuncUtil.compare(object_to_find, item):
                return True
        return False

    @staticmethod
    def find_object_in_list(object_to_find, list_object: list):
        for item in list_object:
            if DrbQueryFuncUtil.compare(object_to_find, item):
                return item
        return None

    @staticmethod
    def append_if_object_not_in(object_to_append, list_object: list):
        if not DrbQueryFuncUtil.is_object_in_list(object_to_append,
                                                  list_object):
            list_object.append(object_to_append)

    @staticmethod
    def append_or_extend_if_not_in(object_to_append, list_object: list):
        if isinstance(object_to_append, list):
            for item in object_to_append:
                DrbQueryFuncUtil.append_if_object_not_in(item, list_object)
        else:
            DrbQueryFuncUtil.append_if_object_not_in(object_to_append,
                                                     list_object)

    @staticmethod
    def remove_duplicate(list_object: list):
        result = []
        for item in list_object:
            item_found = DrbQueryFuncUtil.find_object_in_list(item,
                                                              result)
            if item_found is None:
                result.append(item)
        return result

    @staticmethod
    def intersect(object_to_append, list_object: list):
        if list_object is None:
            if isinstance(object_to_append, list):
                return DrbQueryFuncUtil.remove_duplicate(object_to_append)
            else:
                return [object_to_append]

        result = []
        object_to_append = DrbQueryFuncUtil.remove_duplicate(object_to_append)
        if isinstance(object_to_append, list):
            for item in object_to_append:
                item_found = DrbQueryFuncUtil.find_object_in_list(item,
                                                                  list_object)
                if item_found is not None:
                    result.append(item_found)
        elif DrbQueryFuncUtil.is_object_in_list(object_to_append, list_object):
            result.append(object_to_append)
        return result

    @staticmethod
    def except_op(object_to_append, list_object: list):
        if list_object is None:
            if isinstance(object_to_append, list):
                return DrbQueryFuncUtil.remove_duplicate(object_to_append)
            else:
                return [object_to_append]

        if isinstance(object_to_append, list):
            for item in object_to_append:
                item_found = DrbQueryFuncUtil.find_object_in_list(item,
                                                                  list_object)
                if item_found is not None:
                    list_object.remove(item_found)
        else:
            item_found = DrbQueryFuncUtil.find_object_in_list(object_to_append,
                                                              list_object)
            if item_found is not None:
                list_object.remove(object_to_append)
        return list_object

    @staticmethod
    def union(object_to_append, list_object: list):
        if list_object is None:
            if isinstance(object_to_append, list):
                return DrbQueryFuncUtil.remove_duplicate(object_to_append)
            else:
                return [object_to_append]
        if isinstance(object_to_append, list):
            for item in object_to_append:
                item_found = DrbQueryFuncUtil.find_object_in_list(item,
                                                                  list_object)
                if item_found is None:
                    list_object.append(item)
        else:
            item_found = DrbQueryFuncUtil.find_object_in_list(object_to_append,
                                                              list_object)
            if item_found is None:
                list_object.append(object_to_append)
        return list_object

    @staticmethod
    def append_or_extend(object_to_append, list_object: list):
        if isinstance(object_to_append, list):
            for item in object_to_append:
                list_object.append(item)
        else:
            list_object.append(object_to_append)

    @staticmethod
    def compare_drb(node1: DrbNode, node2: DrbNode) -> bool:
        if node1 == node2:
            if len(node1) == len(node2):
                for index_child in range(len(node1)):
                    if not DrbQueryFuncUtil.compare_drb(node1[index_child],
                                                        node2[index_child]):
                        return False
                return True
        return False

    @staticmethod
    def compare(node1, node2) -> bool:
        if type(node1) != type(node2):
            return False

        if isinstance(node1, list):
            if len(node1) != len(node2):
                return False
            for item_list1, item_list2 in zip(node1, node2):
                if not DrbQueryFuncUtil.compare(item_list1, item_list2):
                    return False
            return True

        if isinstance(node1, DynamicContext):
            return DrbQueryFuncUtil.compare(node1.node, node2.node)
        elif isinstance(node1, DrbNode):
            return DrbQueryFuncUtil.compare_drb(node1, node2)
        else:
            return node1 == node2

    @staticmethod
    def split_namespace_name(nc_name):
        type_var_split = nc_name.split(':')

        if len(type_var_split) > 1:
            return type_var_split[0], type_var_split[1]
        else:
            return None, type_var_split[0]

    @staticmethod
    def replace_group_regex_to_python(patttern):
        previous = ' '
        in_group = False
        value_group = ''
        result = ''

        for char in patttern:

            if in_group:
                if char.isnumeric():
                    value_group = value_group + char
                else:
                    in_group = False
                    if len(value_group) != 0:
                        result = result + r'\g<' + value_group + '>'
                    else:
                        raise DynamicException(ErrorXQUERY.FORX0004,
                                               "Error in replace pattern:" +
                                               patttern)
            if not in_group:
                if char == '$':
                    if previous != '\\':
                        in_group = True
                        value_group = ''
                    else:
                        result = result + char
                else:
                    result = result + char

        if in_group:
            if len(value_group) != 0:
                result = result + r'\g<' + value_group + '>'
            else:
                raise DynamicException(ErrorXQUERY.FORX0004,
                                       "Error in replace pattern:" +
                                       patttern)

        return result



import html
import re
import unittest

from drb import DrbNode

from drb_xquery import DrbXQuery
from drb_xquery.drb_xquery_res_to_string import XQueryResToString


class XQueryTest:

    def __init__(self, node: DrbNode):
        self.node = node

        self.name = node.get_attribute("name")
        self.query = self.node['query']
        try:
            self.expected_result = self.node['result']
        except Exception as Error:
            self.expected_result = None

        self.dynamicError = False
        try:
            attr_value = self.query.get_attribute("dynamicError")
            if attr_value is not None and str(attr_value).lower() == "true":
                self.dynamicError = True
        except Exception as Error:
            pass
        self.staticError = False
        try:
            attr_value = self.query.get_attribute("staticError")
            if attr_value is not None and str(attr_value).lower() == "true":
                self.staticError = True
        except Exception as Error:
            pass

    @staticmethod
    def remove_blank(str_result):
        # Remove blank for simplify the compare operation
        str_result = str(str_result).lower()
        str_result = str_result.replace('\n', '')
        str_result = str_result.strip()
        str_result = re.sub('\\s+', ' ', str_result)
        str_result = str_result.replace(', ', ',')
        str_result = str_result.replace('> ', '>')
        str_result = str_result.replace(' />', '/>')

        return str_result

    @staticmethod
    def compare_result_and_expected(expected_str,
                                    result_string):
        expected_str = XQueryTest.remove_blank(expected_str)
        result_string = XQueryTest.remove_blank(result_string)

        if expected_str == result_string:
            return True

        # import difflib
        # out = list(difflib.Differ().compare(expected_str, result_string))
        # if out:
        #     for line in out:
        #         print(line)
        return False

    def run_test(self, testClass: unittest.TestCase):
        try:
            query_string = self.query.value
            query_string = html.unescape(query_string)

            query = DrbXQuery(query_string)
            result = query.execute(None)
        except Exception as error_query:
            if self.dynamicError or self.staticError:
                return True
            else:
                raise Exception(error_query,
                                'Error raise  in ' + str(self.name))
        if result is None:
            if self.expected_result is None:
                return True
            else:
                return False

        if self.dynamicError or self.staticError:
            print("Test OK but error is waited")
            return False

        if not isinstance(result, list):
            result = [result]

        result_string = None
        for item in result:
            result_string = XQueryResToString.add_item_to_result(
                result_string, item,
                context=query.static_context)

        if result_string is None:
            result_string = ''

        if self.expected_result is None or self.expected_result.value is None:
            expected_result = ''
        else:
            expected_result = self.expected_result.value

        if XQueryTest.compare_result_and_expected(expected_result,
                                                  result_string):
            return True

        testClass.assertTrue(False, self.name
                             + ' result are not as expected\n'
                             + result_string + '\n != \n' +
                             expected_result + '\n')



from enum import Enum
from typing import Optional, Any

from drb import DrbNode

from .execptions import ErrorXQUERY, DynamicException, \
    StaticException


class NavigationStep(Enum):
    CHILD = 1
    DESCENDANT_OR_SELF = 2
    DESCENDANT = 3


class AxeNode:
    def __init__(self, navigation_step=NavigationStep.CHILD):
        self.namespace = None
        self.name = None
        self.is_attribute = False
        self.is_node = True
        self.is_wildcard = False
        self.is_parent = False
        self.navigation_step = navigation_step


class NamespaceMap:

    predefined_namespace = {
        'xmlns': "http://www.w3.org/2000/xmlns/",
        'xml': "http://www.w3.org/XML/1998/namespace",
        'fn': "http://www.w3.org/2005/xpath-functions",
        'xs': "http://www.w3.org/2001/XMLSchema",
        'xsi': "http://www.w3.org/2001/XMLSchema-instance",
        'xdt': "http://www.w3.org/2003/11/xpath-datatypes",
        'local': "http://www.w3.org/2005/xquery-local-functions",
        'drb': "http://www.gael.fr/drb",
        'math': "http://exslt.org/math"
    }

    def __init__(self):
        self.namespace_map = {}
        self.namespace_prefix_map = {}

    @staticmethod
    def check_is_predefined_namespace(prefix: str, namespace_full: str):
        if prefix in NamespaceMap.predefined_namespace.keys():
            raise StaticException(
                ErrorXQUERY.XQST0070, "Prefix namespace "
                                      + prefix + " is predefined and cant be "
                                                 "re declared.")
        if namespace_full in NamespaceMap.predefined_namespace.values():
            raise StaticException(
                ErrorXQUERY.XQST0070, "Namespace "
                                      + namespace_full +
                                      " is predefined and cant be "
                                      "re declared.")

        return False

    def check_duplicate_namespace(self, prefix: str, namespace_full: str):
        if prefix in self.namespace_map.keys():
            raise StaticException(
                ErrorXQUERY.XQST0033, "Prefix namespace "
                                      + prefix + " already declared.")

        return False

    def add_namespace(self, prefix: str, namespace_full: str):
        self.check_duplicate_namespace(prefix, namespace_full)
        self.namespace_map[prefix] = namespace_full
        self.namespace_prefix_map[namespace_full] = prefix

    def get_namespace_full_name(self, prefix: str):
        if prefix in NamespaceMap.predefined_namespace.keys():
            return NamespaceMap.predefined_namespace[prefix]
        if prefix in self.namespace_map.keys():
            return self.namespace_map[prefix]
        return ''

    def get_namespace_prefix_name(self, fullname: str):
        if fullname in NamespaceMap.predefined_namespace.values():
            # TODO Optimise
            for key, full_key in NamespaceMap.predefined_namespace.items():
                if full_key == fullname:
                    return key

        if fullname in self.namespace_prefix_map.keys():
            return self.namespace_prefix_map[fullname]
        return ''

    def clone(self, other):
        self.namespace_map = other.namespace_map.copy()
        self.namespace_prefix_map = other.namespace_prefix_map.copy()


class StaticContext:
    def __init__(self):
        self.list_func = {}
        self.in_scope_variables = {}
        self.namespace_map = {}
        self.namespace_prefix_map = {}
        self.namespace_func = {}
        self.namespace_default_func = ''
        self.namespace_default_elt = ''

        self.name_space_map = NamespaceMap()

        self.namespace_builtin_func = {}

        self.imported_module = {}
        self.is_in_predicate = False

    def clone(self, other):
        if isinstance(other, StaticContext):
            self.namespace_map = other.namespace_map.copy()
            self.namespace_prefix_map = other.namespace_prefix_map.copy()
            self.namespace_func = other.namespace_func.copy()
            self.namespace_builtin_func = other.namespace_func.copy()

            self.name_space_map.clone(other.name_space_map)
            self.imported_module = other.imported_module.copy()

    def get_key_var(self, key):
        from drb_xquery.drb_xquery_utils import DrbQueryFuncUtil
        namespace_prefix, name_var = DrbQueryFuncUtil.split_namespace_name(key)
        namespace_full = None
        if namespace_prefix is not None:
            namespace_full = self.get_namespace_full_name(namespace_prefix)

        if namespace_full is not None and len(namespace_full) > 0:
            return namespace_full, name_var

        else:
            return None, name_var

    def add_var(self, key: str, var):
        key = self.get_key_var(key)

        self.in_scope_variables[key] = var

    def get_var(self, key: str):
        key_tuple = self.get_key_var(key)

        if key_tuple not in self.in_scope_variables.keys():
            raise DynamicException(ErrorXQUERY.XPST0008, "Variable : " +
                                   str(key) + " unknown.")
        return self.in_scope_variables[key_tuple]

    def add_namespace(self, prefix: str, namespace_full: str):
        NamespaceMap.check_is_predefined_namespace(prefix, namespace_full)
        return self.name_space_map.add_namespace(prefix, namespace_full)

    def get_namespace_full_name(self, prefix: str):
        return self.name_space_map.get_namespace_full_name(prefix)

    def get_namespace_prefix_name(self, fullname: str):
        return self.name_space_map.get_namespace_prefix_name(fullname)

    def set_func_list_namespace(self, prefix: str, func_list: dict):
        self.namespace_func[prefix] = func_list

    def get_func_list_namespace(self, prefix: str):
        if prefix in self.namespace_func.keys():
            return self.namespace_func[prefix]
        return None

    def get_func_namespace(self, prefix: str, func_name: str):
        if prefix in self.namespace_func.keys():
            if func_name in self.namespace_func[prefix].keys():
                return self.namespace_func[prefix][func_name]
        return None

    # def add_func_namespace(self, prefix: str, func_name: str, func):
    #     if prefix in self.namespace_func.keys():
    #         self.namespace_func[prefix][func_name] = func
    #     else:
    #         self.namespace_func[prefix] = {}
    #         self.namespace_func[prefix][func_name] = func

    def add_func_build_in(self, func_name: str, func_exp,
                          specific_context=None):
        ns, func_name = self.get_key_var(func_name)
        from drb_xquery.drb_xquery_variable import XqueryBuildIn
        self.namespace_builtin_func[ns, func_name] = XqueryBuildIn(
            func_name,
            func_exp,
            namespace=ns,
            specific_context=specific_context
        )

    def get_func_build_in(self, func_name: str):
        ns, func_name = self.get_key_var(func_name)
        if (ns, func_name) in self.namespace_builtin_func.keys():
            return self.namespace_builtin_func[ns, func_name]
        return None


class DynamicContext:

    def __init__(self, node, position=0, value=None, parent=None,
                 attribute=False,
                 size_context=1,
                 namespace=None,
                 dynamic_context_prev=None):
        self.node = node
        self._value = value
        self._name = None

        if not isinstance(self.node, DrbNode) and value is None:
            self._value = self.node
        self._namespace = namespace
        self._parent = parent
        self.position = position
        self.size_context = size_context
        self.attribute = attribute
        self.name_space_map = NamespaceMap()

        if dynamic_context_prev is None:
            self.namespace_default_elt = ''
        else:
            self.namespace_default_elt = \
                dynamic_context_prev.namespace_default_elt
            self.name_space_map.clone(dynamic_context_prev.name_space_map)

    @property
    def name(self) -> str:
        if self._name is None:
            if isinstance(self.node, DrbNode):
                self._name = self.node.name
            elif isinstance(self.node, tuple):
                if len(self.node) == 2:
                    self._name, self._namespace = self.node
            else:
                self._name = self.node
        return self._name

    @property
    def value(self) -> Optional[Any]:
        if isinstance(self.node, DrbNode):
            return self.node.value
        return self._value

    @property
    def namespace_uri(self) -> Optional[str]:
        if isinstance(self.node, DrbNode):
            self._namespace = self.node.namespace_uri
        elif self._namespace is None and self._name is None:
            if isinstance(self.node, tuple):
                if len(self.node) == 2:
                    self._name, self._namespace = self.node
        return self._namespace

    def is_node(self):
        return isinstance(self.node, DrbNode)

    def is_attribute(self):
        return self.attribute

    @property
    def parent(self) -> Optional[DrbNode]:
        if isinstance(self.node, DrbNode):
            return self.node.parent
        return self._parent

    def add_namespace(self, static_context: StaticContext,
                      prefix: str, namespace_full: str):
        if prefix == 'xml' and \
                NamespaceMap.predefined_namespace[prefix] == namespace_full:
            return
        NamespaceMap.check_is_predefined_namespace(prefix, namespace_full)

        static_context.name_space_map.check_duplicate_namespace(prefix,
                                                                namespace_full)
        return self.name_space_map.add_namespace(prefix, namespace_full)

    def get_namespace_full_name(self, prefix: str):
        return self.name_space_map.get_namespace_full_name(prefix)

    def get_namespace_prefix_name(self, fullname: str):
        return self.name_space_map.get_namespace_prefix_name(fullname)

    def __eq__(self, other):
        if self.is_node():
            return self.node == other.node

        return self.name == other.value and self.value == other.value



import html
import unittest

from drb import DrbNode
from drb_impl_file import DrbFileFactory

from drb_impl_xml import XmlNodeFactory
from drb_xquery import DrbXQuery
from pathlib import Path

from drb_xquery.drb_xquery_res_to_string import XQueryResToString
from drb_xquery.drb_xquery_utils import DrbQueryFuncUtil
from tests.XQueryTest import XQueryTest

X_QUERY_PATH_FILES = 'tests/files/XQTS_1_0_3'


class XQueryTestXQTS:

    def import_var(self, node_input: DrbNode):
        variable_name = node_input.get_attribute('variable')

        if node_input.value != 'emptydoc':
            xml_path = X_QUERY_PATH_FILES + '/TestSources/' + \
                       node_input.value + ".xml"
            node_file = DrbFileFactory().create(xml_path)
            node_var = XmlNodeFactory().create(node_file)
            self.vars[variable_name] = node_var
        else:
            self.vars[variable_name] = None

    def import_var_uri(self, node_input: DrbNode):
        if node_input.value != 'emptydoc':
            xml_path = X_QUERY_PATH_FILES + '/TestSources/' + \
                       node_input.value + ".xml"
            variable_name = node_input.get_attribute('variable')
            self.vars[variable_name] = xml_path

    def import_var_xq(self, node_input: DrbNode, path_file):

        variable_name = node_input.get_attribute('variable')
        name_xq = node_input.get_attribute('name')

        try:
            txt_query = Path(X_QUERY_PATH_FILES + '/Queries/XQuery/' +
                             path_file + '/' + name_xq + '.xq').read_text()
            query = DrbXQuery(txt_query)
            result = query.execute(None)

            if isinstance(result, list):
                result = result[0]
            self.vars[variable_name] = result
        except Exception as error_query:
            raise Exception(error_query,
                            'Error not expected raise  in ' +
                            str(self.name) +
                            " When get var " + variable_name)

    def __init__(self, node: DrbNode):
        self.node = node
        self.name = self.node.get_attribute("name")
        self.expected_result = []
        self.expected_xml = None
        self.vars = {}
        self.node_begin = None
        self.error = None

    def run_test(self, testClass: unittest.TestCase):

        path_file = self.node.get_attribute("FilePath")
        txt_query = Path(X_QUERY_PATH_FILES + '/Queries/XQuery/' +
                         path_file + '/' + self.name + '.xq').read_text()
        self.query = txt_query

        if self.node.has_child('output-file'):
            nodes_res = self.node['output-file', :]
            for node_res in nodes_res:
                file_res = node_res.value
                file_path = X_QUERY_PATH_FILES + '/ExpectedTestResults/'
                file_path += path_file + '/' + file_res
                if file_res.endswith('.xml'):
                    try:
                        self.expected_xml = \
                            XmlNodeFactory().create(source=file_path)
                    except Exception:
                        pass
                self.expected_result.append(Path(file_path).read_text())

        if self.node.has_child('input-query'):
            nodes_var = self.node['input-query', :]
            for node_var in nodes_var:
                self.import_var_xq(node_var, path_file)

        if self.node.has_child('input-file'):
            nodes_res = self.node['input-file', :]
            for node_input in nodes_res:
                self.import_var(node_input)

        if self.node.has_child('input-URI'):
            node_input = self.node['input-URI']
            self.import_var_uri(node_input)

        if self.node.has_child('expected-error'):
            self.error = self.node['expected-error'].value

        try:
            query_string = self.query
            # query_string = html.unescape(query_string)

            query = DrbXQuery(query_string)
            result = query.execute(self.node_begin, external_var=self.vars)
        except Exception as error_query:
            if self.error:
                return True
            else:
                raise Exception(error_query,
                                'Error not expected raise  in ' +
                                str(self.name))
        if result is None:
            if len(self.expected_result) == 0:
                return True
            else:
                return False

        if not isinstance(result, list):
            result = [result]

        if len(result) == 1 and self.expected_xml is not None:
            node = DrbQueryFuncUtil.get_node(result)
            if DrbQueryFuncUtil.compare_drb(self.expected_xml, node):
                return True

        result_string = None
        for item in result:
            result_string = XQueryResToString.add_item_to_result(
                result_string, item,
                separator=' ',
                context=query.static_context,
                float_format_g=False)

        if result_string is None:
            result_string = ''

        if self.error and len(self.expected_result) == 0:
            print("Test OK but error is waited")
            testClass.assertTrue(False, self.name + ' Error ' + self.error +
                                 " is expected, but not error raised " +
                                 " result is :\n" +
                                 result_string)

        if len(self.expected_result) == 0:
            if result_string == '':
                return True

        for expected_result_possible in self.expected_result:
            if XQueryTest.compare_result_and_expected(
                    expected_result_possible,
                    result_string):

                return True

        testClass.assertTrue(False, self.name +
                             ' result are not as expected\n' +
                             "'" + result_string + "'" +
                             '\n != \n' +
                             "'" + expected_result_possible + "'")



import os
import unittest
from pathlib import Path

from drb_impl_file import DrbFileFactory
from drb_impl_xml import XmlNodeFactory

from tests.XQueryTestXQTS import XQueryTestXQTS


class TestDrbXQueryXQTS(unittest.TestCase):
    current_path = Path(os.path.dirname(os.path.realpath(__file__)))

    xml_path = current_path / "files"

    def test_xquery_xqts_func_abs(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/ABSFunc.xml")
        #
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_analyse_string(self):

        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/AnalyzeString.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_direct_con_elt(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/DirectConElem.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_FLWOR(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/FLWORExprSI.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_substring(self):
        xml_file = str(
            self.xml_path / "XQTS_1_0_3/cat/functx-fn-substring.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_substring_after(self):
        xml_file = str(
            self.xml_path / "XQTS_1_0_3/cat/functx-fn-substring-after.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_substring_before(self):
        xml_file = str(
            self.xml_path / "XQTS_1_0_3/cat/functx-fn-substring-before.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_sum(self):
        xml_file = str(
            self.xml_path / "XQTS_1_0_3/cat/functx-fn-sum.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_filter_expr(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/FilterExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_for_exp_wtih(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/ForExprWith.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_import(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/ModuleImport.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_predicates(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/Predicates.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_range(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/RangeExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_return(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/ReturnExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_function_prolog(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/FunctionProlog.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_function_paren(self):
        xml_file = str(self.xml_path / "XQTS_1_0_3/cat/ParenExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_module_comma_op(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/commaOp.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_call_exp(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/FunctionCallExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_lower_cae(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-lower-case.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_upper_cae(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-upper-case.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_reverse(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-reverse.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_in_seq_count_func(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/SeqCountFunc.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_count(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-count.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_distinct_values(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-distinct-values.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_compare(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-compare.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_concat(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-concat.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_function_contains(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-contains.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_compare_function(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/CompareFunction.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_cond_exp(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/CondExpr.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_function_boolean(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-boolean.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_function_not(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-not.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_let_expr(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/LetExprWith.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_deep_equal(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-deep-equal.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_seq_deep_equal(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/SeqDeepEqualFunc.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_empty(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-empty.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_ends_with(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-ends-with.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_exists(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-exists.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_floor(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-floor.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_false(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-false.xml")
        self.from_xml_xqts(xml_file)

    # def test_xquery_xqts_fn_last(self):
    #     xml_file = str(self.xml_path /
    #                    "XQTS_1_0_3/cat/functx-fn-last.xml")
    #     self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_max(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-max.xml")
        self.from_xml_xqts(xml_file)

    def test_xquery_xqts_fn_min(self):
        xml_file = str(self.xml_path /
                       "XQTS_1_0_3/cat/functx-fn-min.xml")
        self.from_xml_xqts(xml_file)

    # def test_xquery_xqts_fn_name(self):
    #     xml_file = str(self.xml_path /
    #                    "XQTS_1_0_3/cat/functx-fn-name.xml")
    #     self.from_xml_xqts(xml_file)
    # def test_xquery_xqts_fn_in_progess(self):
    #     xml_file = str(self.xml_path /
    #                     "XQTS_1_0_3/cat/in_progress.xml")
    #     self.from_xml_xqts(xml_file)

    def from_xml_xqts(self, xml_path):
        node_file = DrbFileFactory().create(xml_path)
        node = XmlNodeFactory().create(node_file)

        document = node['test-group']

        for child in document['test-case', :]:
            current_test_node = child
            test = XQueryTestXQTS(current_test_node)
            with self.subTest(test.name):
                self.assertTrue(test.run_test(self), 'error in test '
                                + test.name)



import os
import unittest
from pathlib import Path

from drb_impl_file import DrbFileFactory
from drb_impl_xml import XmlNodeFactory

from tests.XQueryTest import XQueryTest


class TestDrbXQueryFunc(unittest.TestCase):
    current_path = Path(os.path.dirname(os.path.realpath(__file__)))

    xml_path = current_path / "files"

    def test_xquery_gael(self):
        xml_file = str(self.xml_path / "test_gael.xml")
        self.from_xml_node(xml_file)

    def test_xquery_operator(self):
        xml_file = str(self.xml_path / "xquery-gael-operators.xml")
        self.from_xml_node(xml_file)

    def test_xquery_doc(self):
        xml_file = str(self.xml_path / "xquery-gael-doc.xml")
        self.from_xml_node(xml_file)

    def test_xquery_functions(self):
        xml_file = str(self.xml_path / "xquery-gael-functions.xml")
        self.from_xml_node(xml_file)

    def test_xquery_queries(self):
        xml_file = str(self.xml_path / "xquery-gael-queries.xml")
        self.from_xml_node(xml_file)

    def test_xquery_module(self):
        xml_file = str(self.xml_path / "xquery-gael-module.xml")
        self.from_xml_node(xml_file)
    #
    # def test_xquery_in_progress(self):
    #     xml_file = str(self.xml_path / "test.xml")
    #     self.from_xml_node(xml_file)

    def from_xml_node(self, xml_path):
        node_file = DrbFileFactory().create(xml_path)
        node = XmlNodeFactory().create(node_file)

        document = node[0]

        for child in document.children:
            current_test_node = child
            test = XQueryTest(current_test_node)
            with self.subTest(test.name):
                self.assertTrue(test.run_test(self), 'error in test '
                                + test.name)
        node.close()
        node_file.close()



# Generated from XQueryParser.g4 by ANTLR 4.9.2
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .XQueryParser import XQueryParser
else:
    from XQueryParser import XQueryParser

# This class defines a complete generic visitor for a parse tree produced by XQueryParser.

class XQueryParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by XQueryParser#module.
    def visitModule(self, ctx:XQueryParser.ModuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#xqDocComment.
    def visitXqDocComment(self, ctx:XQueryParser.XqDocCommentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#versionDecl.
    def visitVersionDecl(self, ctx:XQueryParser.VersionDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mainModule.
    def visitMainModule(self, ctx:XQueryParser.MainModuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#queryBody.
    def visitQueryBody(self, ctx:XQueryParser.QueryBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#libraryModule.
    def visitLibraryModule(self, ctx:XQueryParser.LibraryModuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#moduleDecl.
    def visitModuleDecl(self, ctx:XQueryParser.ModuleDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#prolog.
    def visitProlog(self, ctx:XQueryParser.PrologContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#defaultNamespaceDecl.
    def visitDefaultNamespaceDecl(self, ctx:XQueryParser.DefaultNamespaceDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#setter.
    def visitSetter(self, ctx:XQueryParser.SetterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#boundarySpaceDecl.
    def visitBoundarySpaceDecl(self, ctx:XQueryParser.BoundarySpaceDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#defaultCollationDecl.
    def visitDefaultCollationDecl(self, ctx:XQueryParser.DefaultCollationDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#baseURIDecl.
    def visitBaseURIDecl(self, ctx:XQueryParser.BaseURIDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#constructionDecl.
    def visitConstructionDecl(self, ctx:XQueryParser.ConstructionDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#orderingModeDecl.
    def visitOrderingModeDecl(self, ctx:XQueryParser.OrderingModeDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#emptyOrderDecl.
    def visitEmptyOrderDecl(self, ctx:XQueryParser.EmptyOrderDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#copyNamespacesDecl.
    def visitCopyNamespacesDecl(self, ctx:XQueryParser.CopyNamespacesDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#preserveMode.
    def visitPreserveMode(self, ctx:XQueryParser.PreserveModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#inheritMode.
    def visitInheritMode(self, ctx:XQueryParser.InheritModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#decimalFormatDecl.
    def visitDecimalFormatDecl(self, ctx:XQueryParser.DecimalFormatDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#schemaImport.
    def visitSchemaImport(self, ctx:XQueryParser.SchemaImportContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#schemaPrefix.
    def visitSchemaPrefix(self, ctx:XQueryParser.SchemaPrefixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#moduleImport.
    def visitModuleImport(self, ctx:XQueryParser.ModuleImportContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#namespaceDecl.
    def visitNamespaceDecl(self, ctx:XQueryParser.NamespaceDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#varDecl.
    def visitVarDecl(self, ctx:XQueryParser.VarDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#varValue.
    def visitVarValue(self, ctx:XQueryParser.VarValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#varDefaultValue.
    def visitVarDefaultValue(self, ctx:XQueryParser.VarDefaultValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#contextItemDecl.
    def visitContextItemDecl(self, ctx:XQueryParser.ContextItemDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionDecl.
    def visitFunctionDecl(self, ctx:XQueryParser.FunctionDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionParams.
    def visitFunctionParams(self, ctx:XQueryParser.FunctionParamsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionParam.
    def visitFunctionParam(self, ctx:XQueryParser.FunctionParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#annotations.
    def visitAnnotations(self, ctx:XQueryParser.AnnotationsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#annotation.
    def visitAnnotation(self, ctx:XQueryParser.AnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#annotList.
    def visitAnnotList(self, ctx:XQueryParser.AnnotListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#annotationParam.
    def visitAnnotationParam(self, ctx:XQueryParser.AnnotationParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionReturn.
    def visitFunctionReturn(self, ctx:XQueryParser.FunctionReturnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#optionDecl.
    def visitOptionDecl(self, ctx:XQueryParser.OptionDeclContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#expr.
    def visitExpr(self, ctx:XQueryParser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#exprSingle.
    def visitExprSingle(self, ctx:XQueryParser.ExprSingleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#flworExpr.
    def visitFlworExpr(self, ctx:XQueryParser.FlworExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#initialClause.
    def visitInitialClause(self, ctx:XQueryParser.InitialClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#intermediateClause.
    def visitIntermediateClause(self, ctx:XQueryParser.IntermediateClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#forClause.
    def visitForClause(self, ctx:XQueryParser.ForClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#forBinding.
    def visitForBinding(self, ctx:XQueryParser.ForBindingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#allowingEmpty.
    def visitAllowingEmpty(self, ctx:XQueryParser.AllowingEmptyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#positionalVar.
    def visitPositionalVar(self, ctx:XQueryParser.PositionalVarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#letClause.
    def visitLetClause(self, ctx:XQueryParser.LetClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#letBinding.
    def visitLetBinding(self, ctx:XQueryParser.LetBindingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#windowClause.
    def visitWindowClause(self, ctx:XQueryParser.WindowClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#tumblingWindowClause.
    def visitTumblingWindowClause(self, ctx:XQueryParser.TumblingWindowClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#slidingWindowClause.
    def visitSlidingWindowClause(self, ctx:XQueryParser.SlidingWindowClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#windowStartCondition.
    def visitWindowStartCondition(self, ctx:XQueryParser.WindowStartConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#windowEndCondition.
    def visitWindowEndCondition(self, ctx:XQueryParser.WindowEndConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#windowVars.
    def visitWindowVars(self, ctx:XQueryParser.WindowVarsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#countClause.
    def visitCountClause(self, ctx:XQueryParser.CountClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#whereClause.
    def visitWhereClause(self, ctx:XQueryParser.WhereClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#groupByClause.
    def visitGroupByClause(self, ctx:XQueryParser.GroupByClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#groupingSpecList.
    def visitGroupingSpecList(self, ctx:XQueryParser.GroupingSpecListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#groupingSpec.
    def visitGroupingSpec(self, ctx:XQueryParser.GroupingSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#orderByClause.
    def visitOrderByClause(self, ctx:XQueryParser.OrderByClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#orderSpec.
    def visitOrderSpec(self, ctx:XQueryParser.OrderSpecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#returnClause.
    def visitReturnClause(self, ctx:XQueryParser.ReturnClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#quantifiedExpr.
    def visitQuantifiedExpr(self, ctx:XQueryParser.QuantifiedExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#quantifiedVar.
    def visitQuantifiedVar(self, ctx:XQueryParser.QuantifiedVarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#switchExpr.
    def visitSwitchExpr(self, ctx:XQueryParser.SwitchExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#switchCaseClause.
    def visitSwitchCaseClause(self, ctx:XQueryParser.SwitchCaseClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#switchCaseOperand.
    def visitSwitchCaseOperand(self, ctx:XQueryParser.SwitchCaseOperandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typeswitchExpr.
    def visitTypeswitchExpr(self, ctx:XQueryParser.TypeswitchExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#caseClause.
    def visitCaseClause(self, ctx:XQueryParser.CaseClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#sequenceUnionType.
    def visitSequenceUnionType(self, ctx:XQueryParser.SequenceUnionTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#ifExpr.
    def visitIfExpr(self, ctx:XQueryParser.IfExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#tryCatchExpr.
    def visitTryCatchExpr(self, ctx:XQueryParser.TryCatchExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#tryClause.
    def visitTryClause(self, ctx:XQueryParser.TryClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#enclosedTryTargetExpression.
    def visitEnclosedTryTargetExpression(self, ctx:XQueryParser.EnclosedTryTargetExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#catchClause.
    def visitCatchClause(self, ctx:XQueryParser.CatchClauseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#enclosedExpression.
    def visitEnclosedExpression(self, ctx:XQueryParser.EnclosedExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#catchErrorList.
    def visitCatchErrorList(self, ctx:XQueryParser.CatchErrorListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existUpdateExpr.
    def visitExistUpdateExpr(self, ctx:XQueryParser.ExistUpdateExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existReplaceExpr.
    def visitExistReplaceExpr(self, ctx:XQueryParser.ExistReplaceExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existValueExpr.
    def visitExistValueExpr(self, ctx:XQueryParser.ExistValueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existInsertExpr.
    def visitExistInsertExpr(self, ctx:XQueryParser.ExistInsertExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existDeleteExpr.
    def visitExistDeleteExpr(self, ctx:XQueryParser.ExistDeleteExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#existRenameExpr.
    def visitExistRenameExpr(self, ctx:XQueryParser.ExistRenameExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#orExpr.
    def visitOrExpr(self, ctx:XQueryParser.OrExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#andExpr.
    def visitAndExpr(self, ctx:XQueryParser.AndExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#comparisonExpr.
    def visitComparisonExpr(self, ctx:XQueryParser.ComparisonExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringConcatExpr.
    def visitStringConcatExpr(self, ctx:XQueryParser.StringConcatExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#rangeExpr.
    def visitRangeExpr(self, ctx:XQueryParser.RangeExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#additiveExpr.
    def visitAdditiveExpr(self, ctx:XQueryParser.AdditiveExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#multiplicativeExpr.
    def visitMultiplicativeExpr(self, ctx:XQueryParser.MultiplicativeExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#unionExpr.
    def visitUnionExpr(self, ctx:XQueryParser.UnionExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#intersectExceptExpr.
    def visitIntersectExceptExpr(self, ctx:XQueryParser.IntersectExceptExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#instanceOfExpr.
    def visitInstanceOfExpr(self, ctx:XQueryParser.InstanceOfExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#treatExpr.
    def visitTreatExpr(self, ctx:XQueryParser.TreatExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#castableExpr.
    def visitCastableExpr(self, ctx:XQueryParser.CastableExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#castExpr.
    def visitCastExpr(self, ctx:XQueryParser.CastExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#arrowExpr.
    def visitArrowExpr(self, ctx:XQueryParser.ArrowExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#unaryExpression.
    def visitUnaryExpression(self, ctx:XQueryParser.UnaryExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#valueExpr.
    def visitValueExpr(self, ctx:XQueryParser.ValueExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#generalComp.
    def visitGeneralComp(self, ctx:XQueryParser.GeneralCompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#valueComp.
    def visitValueComp(self, ctx:XQueryParser.ValueCompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#nodeComp.
    def visitNodeComp(self, ctx:XQueryParser.NodeCompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#validateExpr.
    def visitValidateExpr(self, ctx:XQueryParser.ValidateExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#validationMode.
    def visitValidationMode(self, ctx:XQueryParser.ValidationModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#extensionExpr.
    def visitExtensionExpr(self, ctx:XQueryParser.ExtensionExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#simpleMapExpr.
    def visitSimpleMapExpr(self, ctx:XQueryParser.SimpleMapExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#pathExpr.
    def visitPathExpr(self, ctx:XQueryParser.PathExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#relativePathExpr.
    def visitRelativePathExpr(self, ctx:XQueryParser.RelativePathExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stepExpr.
    def visitStepExpr(self, ctx:XQueryParser.StepExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#axisStep.
    def visitAxisStep(self, ctx:XQueryParser.AxisStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#forwardStep.
    def visitForwardStep(self, ctx:XQueryParser.ForwardStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#forwardAxis.
    def visitForwardAxis(self, ctx:XQueryParser.ForwardAxisContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#abbrevForwardStep.
    def visitAbbrevForwardStep(self, ctx:XQueryParser.AbbrevForwardStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#reverseStep.
    def visitReverseStep(self, ctx:XQueryParser.ReverseStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#reverseAxis.
    def visitReverseAxis(self, ctx:XQueryParser.ReverseAxisContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#abbrevReverseStep.
    def visitAbbrevReverseStep(self, ctx:XQueryParser.AbbrevReverseStepContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#nodeTest.
    def visitNodeTest(self, ctx:XQueryParser.NodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#nameTest.
    def visitNameTest(self, ctx:XQueryParser.NameTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#allNames.
    def visitAllNames(self, ctx:XQueryParser.AllNamesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#allWithNS.
    def visitAllWithNS(self, ctx:XQueryParser.AllWithNSContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#allWithLocal.
    def visitAllWithLocal(self, ctx:XQueryParser.AllWithLocalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#postfixExpr.
    def visitPostfixExpr(self, ctx:XQueryParser.PostfixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#argumentList.
    def visitArgumentList(self, ctx:XQueryParser.ArgumentListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#predicateList.
    def visitPredicateList(self, ctx:XQueryParser.PredicateListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#predicate.
    def visitPredicate(self, ctx:XQueryParser.PredicateContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#lookup.
    def visitLookup(self, ctx:XQueryParser.LookupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#keySpecifier.
    def visitKeySpecifier(self, ctx:XQueryParser.KeySpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#arrowFunctionSpecifier.
    def visitArrowFunctionSpecifier(self, ctx:XQueryParser.ArrowFunctionSpecifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#primaryExpr.
    def visitPrimaryExpr(self, ctx:XQueryParser.PrimaryExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#literal.
    def visitLiteral(self, ctx:XQueryParser.LiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#numericLiteral.
    def visitNumericLiteral(self, ctx:XQueryParser.NumericLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#varRef.
    def visitVarRef(self, ctx:XQueryParser.VarRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#varName.
    def visitVarName(self, ctx:XQueryParser.VarNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#parenthesizedExpr.
    def visitParenthesizedExpr(self, ctx:XQueryParser.ParenthesizedExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#contextItemExpr.
    def visitContextItemExpr(self, ctx:XQueryParser.ContextItemExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#orderedExpr.
    def visitOrderedExpr(self, ctx:XQueryParser.OrderedExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#unorderedExpr.
    def visitUnorderedExpr(self, ctx:XQueryParser.UnorderedExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionCall.
    def visitFunctionCall(self, ctx:XQueryParser.FunctionCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#argument.
    def visitArgument(self, ctx:XQueryParser.ArgumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#nodeConstructor.
    def visitNodeConstructor(self, ctx:XQueryParser.NodeConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#directConstructor.
    def visitDirectConstructor(self, ctx:XQueryParser.DirectConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirElemConstructorOpenClose.
    def visitDirElemConstructorOpenClose(self, ctx:XQueryParser.DirElemConstructorOpenCloseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirElemConstructorSingleTag.
    def visitDirElemConstructorSingleTag(self, ctx:XQueryParser.DirElemConstructorSingleTagContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeList.
    def visitDirAttributeList(self, ctx:XQueryParser.DirAttributeListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeValueApos.
    def visitDirAttributeValueApos(self, ctx:XQueryParser.DirAttributeValueAposContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeValueQuot.
    def visitDirAttributeValueQuot(self, ctx:XQueryParser.DirAttributeValueQuotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeValue.
    def visitDirAttributeValue(self, ctx:XQueryParser.DirAttributeValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeContentQuot.
    def visitDirAttributeContentQuot(self, ctx:XQueryParser.DirAttributeContentQuotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirAttributeContentApos.
    def visitDirAttributeContentApos(self, ctx:XQueryParser.DirAttributeContentAposContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#dirElemContent.
    def visitDirElemContent(self, ctx:XQueryParser.DirElemContentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#commonContent.
    def visitCommonContent(self, ctx:XQueryParser.CommonContentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#computedConstructor.
    def visitComputedConstructor(self, ctx:XQueryParser.ComputedConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONConstructor.
    def visitCompMLJSONConstructor(self, ctx:XQueryParser.CompMLJSONConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONArrayConstructor.
    def visitCompMLJSONArrayConstructor(self, ctx:XQueryParser.CompMLJSONArrayConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONObjectConstructor.
    def visitCompMLJSONObjectConstructor(self, ctx:XQueryParser.CompMLJSONObjectConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONNumberConstructor.
    def visitCompMLJSONNumberConstructor(self, ctx:XQueryParser.CompMLJSONNumberConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONBooleanConstructor.
    def visitCompMLJSONBooleanConstructor(self, ctx:XQueryParser.CompMLJSONBooleanConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compMLJSONNullConstructor.
    def visitCompMLJSONNullConstructor(self, ctx:XQueryParser.CompMLJSONNullConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compBinaryConstructor.
    def visitCompBinaryConstructor(self, ctx:XQueryParser.CompBinaryConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compDocConstructor.
    def visitCompDocConstructor(self, ctx:XQueryParser.CompDocConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compElemConstructor.
    def visitCompElemConstructor(self, ctx:XQueryParser.CompElemConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#enclosedContentExpr.
    def visitEnclosedContentExpr(self, ctx:XQueryParser.EnclosedContentExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compAttrConstructor.
    def visitCompAttrConstructor(self, ctx:XQueryParser.CompAttrConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compNamespaceConstructor.
    def visitCompNamespaceConstructor(self, ctx:XQueryParser.CompNamespaceConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#prefix.
    def visitPrefix(self, ctx:XQueryParser.PrefixContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#enclosedPrefixExpr.
    def visitEnclosedPrefixExpr(self, ctx:XQueryParser.EnclosedPrefixExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#enclosedURIExpr.
    def visitEnclosedURIExpr(self, ctx:XQueryParser.EnclosedURIExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compTextConstructor.
    def visitCompTextConstructor(self, ctx:XQueryParser.CompTextConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compCommentConstructor.
    def visitCompCommentConstructor(self, ctx:XQueryParser.CompCommentConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#compPIConstructor.
    def visitCompPIConstructor(self, ctx:XQueryParser.CompPIConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionItemExpr.
    def visitFunctionItemExpr(self, ctx:XQueryParser.FunctionItemExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#namedFunctionRef.
    def visitNamedFunctionRef(self, ctx:XQueryParser.NamedFunctionRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#inlineFunctionRef.
    def visitInlineFunctionRef(self, ctx:XQueryParser.InlineFunctionRefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionBody.
    def visitFunctionBody(self, ctx:XQueryParser.FunctionBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mapConstructor.
    def visitMapConstructor(self, ctx:XQueryParser.MapConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mapConstructorEntry.
    def visitMapConstructorEntry(self, ctx:XQueryParser.MapConstructorEntryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#arrayConstructor.
    def visitArrayConstructor(self, ctx:XQueryParser.ArrayConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#squareArrayConstructor.
    def visitSquareArrayConstructor(self, ctx:XQueryParser.SquareArrayConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#curlyArrayConstructor.
    def visitCurlyArrayConstructor(self, ctx:XQueryParser.CurlyArrayConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringConstructor.
    def visitStringConstructor(self, ctx:XQueryParser.StringConstructorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringConstructorContent.
    def visitStringConstructorContent(self, ctx:XQueryParser.StringConstructorContentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#charNoGrave.
    def visitCharNoGrave(self, ctx:XQueryParser.CharNoGraveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#charNoLBrace.
    def visitCharNoLBrace(self, ctx:XQueryParser.CharNoLBraceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#charNoRBrack.
    def visitCharNoRBrack(self, ctx:XQueryParser.CharNoRBrackContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringConstructorChars.
    def visitStringConstructorChars(self, ctx:XQueryParser.StringConstructorCharsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringConstructorInterpolation.
    def visitStringConstructorInterpolation(self, ctx:XQueryParser.StringConstructorInterpolationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#unaryLookup.
    def visitUnaryLookup(self, ctx:XQueryParser.UnaryLookupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#singleType.
    def visitSingleType(self, ctx:XQueryParser.SingleTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typeDeclaration.
    def visitTypeDeclaration(self, ctx:XQueryParser.TypeDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#sequenceType.
    def visitSequenceType(self, ctx:XQueryParser.SequenceTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#itemType.
    def visitItemType(self, ctx:XQueryParser.ItemTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#atomicOrUnionType.
    def visitAtomicOrUnionType(self, ctx:XQueryParser.AtomicOrUnionTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#kindTest.
    def visitKindTest(self, ctx:XQueryParser.KindTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#anyKindTest.
    def visitAnyKindTest(self, ctx:XQueryParser.AnyKindTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#binaryNodeTest.
    def visitBinaryNodeTest(self, ctx:XQueryParser.BinaryNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#documentTest.
    def visitDocumentTest(self, ctx:XQueryParser.DocumentTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#textTest.
    def visitTextTest(self, ctx:XQueryParser.TextTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#commentTest.
    def visitCommentTest(self, ctx:XQueryParser.CommentTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#namespaceNodeTest.
    def visitNamespaceNodeTest(self, ctx:XQueryParser.NamespaceNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#piTest.
    def visitPiTest(self, ctx:XQueryParser.PiTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#attributeTest.
    def visitAttributeTest(self, ctx:XQueryParser.AttributeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#attributeNameOrWildcard.
    def visitAttributeNameOrWildcard(self, ctx:XQueryParser.AttributeNameOrWildcardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#schemaAttributeTest.
    def visitSchemaAttributeTest(self, ctx:XQueryParser.SchemaAttributeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#elementTest.
    def visitElementTest(self, ctx:XQueryParser.ElementTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#elementNameOrWildcard.
    def visitElementNameOrWildcard(self, ctx:XQueryParser.ElementNameOrWildcardContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#schemaElementTest.
    def visitSchemaElementTest(self, ctx:XQueryParser.SchemaElementTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#elementDeclaration.
    def visitElementDeclaration(self, ctx:XQueryParser.ElementDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#attributeName.
    def visitAttributeName(self, ctx:XQueryParser.AttributeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#elementName.
    def visitElementName(self, ctx:XQueryParser.ElementNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#simpleTypeName.
    def visitSimpleTypeName(self, ctx:XQueryParser.SimpleTypeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typeName.
    def visitTypeName(self, ctx:XQueryParser.TypeNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionTest.
    def visitFunctionTest(self, ctx:XQueryParser.FunctionTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#anyFunctionTest.
    def visitAnyFunctionTest(self, ctx:XQueryParser.AnyFunctionTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typedFunctionTest.
    def visitTypedFunctionTest(self, ctx:XQueryParser.TypedFunctionTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mapTest.
    def visitMapTest(self, ctx:XQueryParser.MapTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#anyMapTest.
    def visitAnyMapTest(self, ctx:XQueryParser.AnyMapTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typedMapTest.
    def visitTypedMapTest(self, ctx:XQueryParser.TypedMapTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#arrayTest.
    def visitArrayTest(self, ctx:XQueryParser.ArrayTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#anyArrayTest.
    def visitAnyArrayTest(self, ctx:XQueryParser.AnyArrayTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#typedArrayTest.
    def visitTypedArrayTest(self, ctx:XQueryParser.TypedArrayTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#parenthesizedItemTest.
    def visitParenthesizedItemTest(self, ctx:XQueryParser.ParenthesizedItemTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#attributeDeclaration.
    def visitAttributeDeclaration(self, ctx:XQueryParser.AttributeDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlNodeTest.
    def visitMlNodeTest(self, ctx:XQueryParser.MlNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlArrayNodeTest.
    def visitMlArrayNodeTest(self, ctx:XQueryParser.MlArrayNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlObjectNodeTest.
    def visitMlObjectNodeTest(self, ctx:XQueryParser.MlObjectNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlNumberNodeTest.
    def visitMlNumberNodeTest(self, ctx:XQueryParser.MlNumberNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlBooleanNodeTest.
    def visitMlBooleanNodeTest(self, ctx:XQueryParser.MlBooleanNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#mlNullNodeTest.
    def visitMlNullNodeTest(self, ctx:XQueryParser.MlNullNodeTestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#eqName.
    def visitEqName(self, ctx:XQueryParser.EqNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#qName.
    def visitQName(self, ctx:XQueryParser.QNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#ncName.
    def visitNcName(self, ctx:XQueryParser.NcNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#functionName.
    def visitFunctionName(self, ctx:XQueryParser.FunctionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#keyword.
    def visitKeyword(self, ctx:XQueryParser.KeywordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#keywordNotOKForFunction.
    def visitKeywordNotOKForFunction(self, ctx:XQueryParser.KeywordNotOKForFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#keywordOKForFunction.
    def visitKeywordOKForFunction(self, ctx:XQueryParser.KeywordOKForFunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#uriLiteral.
    def visitUriLiteral(self, ctx:XQueryParser.UriLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringLiteralQuot.
    def visitStringLiteralQuot(self, ctx:XQueryParser.StringLiteralQuotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringLiteralApos.
    def visitStringLiteralApos(self, ctx:XQueryParser.StringLiteralAposContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringLiteral.
    def visitStringLiteral(self, ctx:XQueryParser.StringLiteralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringContentQuot.
    def visitStringContentQuot(self, ctx:XQueryParser.StringContentQuotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#stringContentApos.
    def visitStringContentApos(self, ctx:XQueryParser.StringContentAposContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by XQueryParser#noQuotesNoBracesNoAmpNoLAng.
    def visitNoQuotesNoBracesNoAmpNoLAng(self, ctx:XQueryParser.NoQuotesNoBracesNoAmpNoLAngContext):
        return self.visitChildren(ctx)



del XQueryParser


import io

import click
from drb.factory import DrbFactoryResolver

from drb_xquery.drb_xquery_res_to_string import XQueryResToString
from drb_xquery.drb_xquery_utils import DrbQueryFuncUtil
from drb_xquery.drb_xquery_variable import XqueryVariable
from .drb_xquery import DrbXQuery


class OptionEatAll(click.Option):

    def __init__(self, *args, **kwargs):
        self.save_other_options = kwargs.pop('save_other_options', True)
        nargs = kwargs.pop('nargs', -1)
        assert nargs == -1, 'nargs, if set, must be -1 not {}'.format(nargs)
        super(OptionEatAll, self).__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._eat_all_parser = None

    def add_to_parser(self, parser, ctx):

        def parser_process(value, state):
            # method to hook to the parser.process
            done = False
            value = [value]
            if self.save_other_options:
                # grab everything up to the next option
                while state.rargs and not done:
                    for prefix in self._eat_all_parser.prefixes:
                        if state.rargs[0].startswith(prefix):
                            done = True
                    if not done:
                        value.append(state.rargs.pop(0))
            else:
                # grab everything remaining
                value += state.rargs
                state.rargs[:] = []
            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        retval = super(OptionEatAll, self).add_to_parser(parser, ctx)
        for name in self.opts:
            our_parser = parser._long_opt.get(name) or \
                         parser._short_opt.get(name)
            if our_parser:
                self._eat_all_parser = our_parser
                self._previous_parser_process = our_parser.process
                our_parser.process = parser_process
                break
        return retval


@click.command()
@click.option('-s', '--string',
              default='',
              help=' Command line string to be evaluated as XQuery script. '
                   'This parameter cannot be used jointly with -f. '
                   'At least -s or -f is to be set.')
@click.option('-f',
              '--file',
              default='',
              help=' Path to a file containing the sctript to be evaluated. '
                   'This parameter cannot be used jointly '
                   'with -s. At least -f or -s is to be set..')
@click.option('-n',
              '--url-node',
              default='',
              help='Url to a node, in drb meaning, that give '
                   'the context of the query. It can be the path'
                   'of a xml file for example on which the xquery will'
                   'be executed')
@click.option('-V',
              '--verbose',
              is_flag=True)
@click.option('-v',
              '--variable',
              default=[],
              multiple=True,
              cls=OptionEatAll,
              help='Variable define -v <QName> <value> [ as <type>].'
                   'Pushes an external variable in the environment '
                   'prior to parse and evaluate the XQuery script. '
                   'The variable is pushed in the  the environment altough '
                   'it has not been declared has an external variable, '
                   'to provide it to the potential nested XQuerys '
                   '(e.g. a call to evaluate-uri() built-in function). '
                   '<QName> is the qualified name of the variable to declare '
                   'whether <value> is a string to bind as value '
                   'of the variable. '
                   'If <QName> matches a typed external variable '
                   'declared in the script, <value> is converted to that '
                   'type before being bound. '
                   'The trailing "as <type>" is optional and shall follow the '
                   'XQuery sequence type declaration (e.g. as xs:integer+ )')
def drb_xquery_cmd(string,
                   file,
                   variable,
                   url_node,
                   verbose):
    """This command evaluates the XQuery script provided as a string or
        a file. The output of the evaluation is printed out in the
        standard output. The output format may have several forms
        according to the resulting sequence. Basically the resulting
        nodes are output as XML fragments, the attributes not attached
        to nodes are written as in XML but prefixed with '@' symbol and
        finally, the atomic values are printed without decoration,
        according to the XML Schema lexical space definitions. All items
        of the output sequence are comma separated. A '()' result denotes
        the empty sequence.
        ."""

    if len(string) == 0 and len(file) == 0:
        ctx = click.get_current_context()
        ctx.fail("At least option query or query-file must be set")

    if len(string) > 0 and len(file) > 0:
        ctx = click.get_current_context()
        ctx.fail("Query and query url are mutually exclusive")

    resolver = DrbFactoryResolver()
    node_query = None

    if len(file) > 0:
        if verbose:
            print('query_file :')
            print(file)
        node_query = resolver.create(file)
        buffer = node_query.get_impl(io.BufferedIOBase)
        string = buffer.read().decode('utf-8')

    node = None
    if len(url_node) > 0:
        node = resolver.create(url_node)

    variables = {}

    for var in variable:
        var = var.lstrip("(")
        var = var.rstrip(")")

        var_split = var.split(',')
        var = []
        for index in range(len(var_split)):
            var.append(var_split[index].strip(' ').strip("'"))
        if len(var) == 2:
            variables[var[0]] = var[1]
        elif len(var) != 4:
            ctx = click.get_current_context()
            ctx.fail("Typed variables " + str(var) + " is malformed")
        else:
            if var[2].lower() != 'as':
                ctx = click.get_current_context()
                ctx.fail("Typed variables " + str(var) + " is malformed")
            type_var_namespace, type_var = \
                DrbQueryFuncUtil.split_namespace_name(var[3])
            if type_var_namespace is None:
                type_var_namespace = 'xs'
            value = XqueryVariable.get_value_typed(type_var,
                                                   type_var_namespace,
                                                   var[1])
            variables[var[0]] = value

    if verbose:
        print('Query :')
        print(string)

        print('Variables :')
        print(variables)

    drb_query = DrbXQuery(string)
    res = drb_query.execute(node, variables)
    if node:
        node.close()
    if node_query:
        node_query.close()
    first = True

    if isinstance(res, list) and (len(res)) == 0:
        print('()')
    else:
        for item in res:
            if first:
                first = False
            else:
                print(',')
            result_string = None
            result_string = XQueryResToString.add_item_to_result(
                result_string, item,
                context=drb_query.static_context)
            print(result_string)


if __name__ == '__main__':
    drb_xquery_cmd()



import unittest

from click.testing import CliRunner

from drb_xquery import drb_xquery_cmd

OPTION_VAR = '-v'
QUERY_STRING_KEY = '-s'
QUERY_URL_KEY = '-f'


class TestCmd(unittest.TestCase):

    def test_command_with_help(self):
        runner = CliRunner()
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd,
                               ['--help'])

    def test_command_with_query_string(self):
        runner = CliRunner()
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd,
                               [QUERY_STRING_KEY,
                                'if (1 != 0) then '
                                'if (4 != 5) then 1 '
                                'else 2 else 3'])
        self.assertEqual(result.exit_code, 0)
        output = result.output.rstrip('\r')
        output = output.rstrip('\n')

        self.assertEqual(output, '1')

    def test_command_with_query_string_and_var(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'declare variable $x external; '
            'declare variable $y external := 5; '
            '$x +$y', OPTION_VAR, 'x',  '9', '-v', 'y', '2'])

        self.assertEqual(result.exit_code, 0)
        output = result.output.rstrip('\r')
        output = output.rstrip('\n')

        self.assertEqual(output, '11')

    def test_command_with_query_string_and_typed_var(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'declare variable $x external; '
            'declare variable $y external := 5; '
            '$x +$y',
            OPTION_VAR, 'x', '9', 'as', 'xs:integer',
            OPTION_VAR, 'y', '2', 'as', 'xs:integer'])

        self.assertEqual(result.exit_code, 0)
        output = result.output.rstrip('\r')
        output = output.rstrip('\n')

        self.assertEqual(output, '11')

    def test_command_with_query_string_and_typed_var_no_ns(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'declare variable $x external; '
            'declare variable $y external := 5; '
            '$x +$y',
            OPTION_VAR, 'x', '9', 'as', 'integer',
            OPTION_VAR, 'y', '2', 'as', 'integer'])
        self.assertEqual(result.exit_code, 0)
        output = result.output.rstrip('\r')
        output = output.rstrip('\n')

        self.assertEqual(output, '11')

    def test_command_with_query_string_and_bad_type_var(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'declare variable $x external; '
            'declare variable $y external := 5; '
            '$x',
            OPTION_VAR, 'x', '1', 'as', 'xs:integer',
            OPTION_VAR, 'y', 'g', 'as', 'xs:integer'])

        self.assertEqual(result.exit_code, 1)

    def test_command_with_query_string_and_malformed_type_var(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'declare variable $x external; '
            'declare variable $y external := 5; '
            '$x',
            OPTION_VAR, 'x', '1', 'as', 'xs:integer',
            OPTION_VAR, 'y', '3', 'to', 'xs:integer'])

        self.assertIn('malformed', result.output.lower())
        self.assertEqual(result.exit_code, 2)

    def test_command_with_query_file_and_node(self):
        runner = CliRunner()

        # data( /Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or
        # serie='B'][position() < 5][2] / name)
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_URL_KEY,
            './tests/files/test.xq',
            '--url-node',
            './tests/files/MTD_TL.xml'])
        self.assertEqual(result.exit_code, 0)

        self.assertTrue('first_doublon' in result.output)

    def test_command_with_query_string_and_node(self):
        runner = CliRunner()

        # data( /Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or
        # serie='B'][position() < 5][2] / name)
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            ' ./Level-2A_Tile_ID/Test_FLF/elt',
            '--url-node',
            './tests/files/MTD_TL.xml'])
        self.assertEqual(result.exit_code, 0)

        self.assertIn('<elt>This is one</elt>', result.output)
        self.assertIn('<elt>End</elt>', result.output)

    def test_command_with_query_string_and_node_attr(self):
        runner = CliRunner()

        # data( /Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or
        # serie='B'][position() < 5][2] / name)
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            ' ./Level-2A_Tile_ID/Test_FLF/@name_attr',
            '--url-node',
            './tests/files/MTD_TL.xml'])
        self.assertEqual(result.exit_code, 0)

        self.assertIn('@name_attr=test_three', result.output)

    def test_command_with_query_string_and_url_query(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_STRING_KEY,
            'test',
            QUERY_URL_KEY,
            './tests/files/test.xq'])

        self.assertIn('exclusive', result.output.lower())
        self.assertEqual(result.exit_code, 2)

    def test_command_verbose(self):
        runner = CliRunner()

        # data( /Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or
        # serie='B'][position() < 5][2] / name)
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            '--verbose',
            QUERY_URL_KEY,
            './tests/files/test.xq',
            OPTION_VAR,
            'myvar',
            'valeu_var',
            '--url-node',
            './tests/files/MTD_TL.xml'])

        self.assertIn('Level-2A_Tile_ID/Test_FLF', result.output)
        self.assertIn('myvar', result.output)

        self.assertEqual(result.exit_code, 0)

    def test_command_no_verbose(self):
        runner = CliRunner()

        # data( /Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or
        # serie='B'][position() < 5][2] / name)
        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [
            QUERY_URL_KEY,
            './tests/files/test.xq',
            OPTION_VAR,
            'myvar',
            'valeu_var',
            '--url-node',
            './tests/files/MTD_TL.xml'])

        self.assertNotIn('Level-2A_Tile_ID/Test_FLF', result.output)
        self.assertNotIn('myvar', result.output)

        self.assertEqual(result.exit_code, 0)

    def test_command_without_query_string_and_url_query(self):
        runner = CliRunner()

        result = runner.invoke(drb_xquery_cmd.drb_xquery_cmd, [])

        self.assertIn('at least', result.output.lower())
        self.assertEqual(result.exit_code, 2)



import os
import unittest
from decimal import Decimal
from pathlib import Path

from drb_impl_file import DrbFileFactory
from drb_impl_xml import XmlNodeFactory
from drb_xquery.drb_xquery import DrbXQuery


class NodeTest:
    def __init__(self):
        self.value = 'test'


class TestDrbXQuery(unittest.TestCase):
    current_path = Path(os.path.dirname(os.path.realpath(__file__)))

    xml_file = current_path / "files" / "MTD_TL.xml"
    xml_file = str(xml_file)

    def test_xml(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)
        #
        # self.node_file = DrbHttpFactory().create(self.xml_file)
        # self.node = XmlNodeFactory().create(self.node_file)

        print(len(self.node))
        # self.assertEqual(len(self.node), 1)

        self.assertEqual(self.node[0].name, "Level-2A_Tile_ID")
        print(self.node[0]["General_Info"]['L1C_TILE_ID'].value)
        print(self.node[0]["General_Info", 'nc3'].namespace_uri)

    def test_xquery_simple(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID")

        node_result = query.execute(self.node)
        print(node_result[0].value)

        self.assertTrue(node_result[0].name,
                        "Level-2A_Tile_ID")

    def test_xquery_double(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "123E+14")

        node_result = query.execute(self.node)
        self.assertEqual(node_result[0], 123e+14)

    def test_xquery_decimal(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)
        query = DrbXQuery(
            "1.40")

        node_result = query.execute(self.node)
        self.assertEqual(node_result[0], Decimal('1.40'))

    def test_xquery_last(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[myday = 1]/L1C_TILE_ID[last()]")

        node_result = query.execute(self.node)
        print(node_result[0].value)

        self.assertTrue(node_result[0].value,
                        "S2B_OPER_MSI_L1C_TL_VGS4_20210913T120150_"
                        "A023615_T30UWU_N03.03")

    def test_xquery_index(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[myday = 1]/L1C_TILE_ID[2]")

        node_result = query.execute(self.node)

        print(node_result[0].value)
        self.assertTrue(node_result[0].value,
                        "S2B_OPER_MSI_L1C_TL_VGS4_20210913T120150_")

    def test_xquery_string(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[\"A\"]")

        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 6)

    def test_xquery_sequence(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[3, 5]")

        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 2)

        self.assertTrue(node_result[0].name, "second")
        self.assertTrue(node_result[1].name, "first_doublon")

        # query = DrbXQuery(
        #     "/Level-2A_Tile_ID/Test_FLF[serie='A', number='one']")
        # node_result = query.execute(self.node)
        #
        # self.assertEqual(len(node_result), 4)
        #
        # query = DrbXQuery(
        #     "/Level-2A_Tile_ID/Test_FLF[serie='A', 4, 5]")
        # node_result = query.execute(self.node)
        #
        # self.assertEqual(len(node_result), 5)

    def test_xquery_range(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[2 to 4]")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 3)
        self.assertTrue(node_result[0].name, "third")
        self.assertTrue(node_result[2].name, "second_doublon")

    def test_xquery_range_and_index(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[2 to 4, 5]")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 4)
        self.assertTrue(node_result[0].name, "third")
        self.assertTrue(node_result[2].name, "second_doublon")

    def test_xquery_equal(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[myday = 2]/L1C_TILE_ID[1]")

        node_result = query.execute(self.node)

        print(node_result[0].value)
        self.assertTrue(node_result[0].value,
                        "S2B_OPER_MSI_L1C_TL_VGS4_20210913T120150_"
                        "A023615_T30UWU_N03.01_day2")

    def test_xquery_greater(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[myday = 2]/"
            "L1C_TILE_ID[position() > 2]")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 1)

    def test_xquery_wilcard(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/*/SENSING_TIME")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 3)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/*/NOT_EXIST")

        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 0)

    def test_xquery_attribute(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[@month=11]/myday")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].value, "2")

    def test_xquery_namespace(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/nc3:General_Info[@month=11]/myday")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].value, "2")

        query = DrbXQuery(
            "/Level-2A_Tile_ID/nc2:General_Info")

        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 0)

    def test_xquery_before_last(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[myday = 1]/L1C_TILE_ID[last()-1]")

        node_result = query.execute(self.node)
        print(node_result[0].value)

        self.assertTrue(node_result[0].value,
                        "S2B_OPER_MSI_L1C_TL_VGS4_20210913T120150_"
                        "A023615_T30UWU_N03.02")

    def test_xquery_wilcard_attribute(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[@*=11]/myday")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].value, "2")

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info[@*=14]/myday")

        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 0)

    def test_xquery_axe_child(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/child::General_Info/myday")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(len(node_result), 2)
        self.assertEqual(node_result[0].value, "1")
        self.assertEqual(node_result[1].value, "2")

        self.node_file.close()
        self.node.close()

    def test_xquery_axe_attribute(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info/attribute::month")

        node_result = query.execute(self.node)

        print("len :" + str(len(node_result)))
        self.assertEqual(node_result[0].value, "01")
        self.assertEqual(node_result[1].value, "11")

    def test_xquery_axe_following(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info/following::month")
        with self.assertRaises(NotImplementedError):
            query.execute(self.node)

    def test_xquery_axe_not_exist(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info/test::myday")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 0)

        # if we ask an axe that not exist the
        query = DrbXQuery(
            "/Level-2A_Tile_ID/General_Info/myday::test")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 2)

    def test_xquery_and(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            '/Level-2A_Tile_ID/Test_FLF[@name_attr="test_one"]')
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 2)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[serie='B']")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 3)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[serie='B' and @name_attr='test_one']")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].node['name'].value, 'first_doublon')

    def test_xquery_or(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            '/Level-2A_Tile_ID/Test_FLF[@name_attr="test_two"]')
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 2)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[serie='B']")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 3)
        self.assertEqual(node_result[0].node['serie'].value, 'B')

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' or serie='B']")
        node_result = query.execute(self.node)
        self.assertEqual(len(node_result), 4)

        self.assertEqual(node_result[0].node['name'].value, 'second')

    def test_xquery_list_predicate(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' "
            "or serie='B'][last()]")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].node['name'].value, 'third_doublon')

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' "
            "or serie='B'][position() < 5]")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 4)
        self.assertEqual(node_result[0].node['name'].value, 'second')

        query = DrbXQuery(
            "/Level-2A_Tile_ID/Test_FLF[@name_attr='test_two' "
            "or serie='B'][position() < 5][2]")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0].node['name'].value, 'first_doublon')

    def test_xquery_func_range_fix(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("1 to 5")
        node_result = query.execute(self.node)

        self.assertListEqual(node_result, [1, 2, 3, 4, 5])

    def test_xquery_namespace_bis(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("/nc3:Level-2A_Tile_ID")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(len(node_result[0].node['Test_FLF', :]), 6)

    def test_xquery_namespace_ter(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("/nc3:Level-2A_Tile_ID/(nc3:Test_num[1], "
                          "nc3:Test_FLF[1])")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 2)

        self.assertEqual(node_result[1].node['name'].value, 'first')

    def test_xquery_namespace_not_exist(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("/nc3:Level-2A_Tile_ID/(nc3:Test_num[1], "
                          "nc4:Test_FLF[1])")
        node_result = query.execute(self.node)

        self.assertEqual(len(node_result), 1)

        self.assertEqual(node_result[0].node['num_to_test'].value, '3.4')

    def test_xquery_external_var(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("declare variable $x external; $x+1")

        list_var = {'x': 2}

        node_result = query.execute(self.node, list_var)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0], 3)

    def test_xquery_external_var_error(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("declare variable $x external; $x+1")

        list_var = {'y': 2}

        with self.assertRaises(Exception) as error:
            node_result = query.execute(self.node, list_var)
        self.assertIn("XPDY0002", str(error.exception))

    def test_xquery_external_var_default_not_defined(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("declare variable $x external; "
                          "declare variable $y external := 5; $x+$y")

        list_var = {'x': 9}

        node_result = query.execute(self.node, list_var)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0], 14)

    def test_xquery_external_var_default_defined(self):
        self.node_file = DrbFileFactory().create(self.xml_file)
        self.node = XmlNodeFactory().create(self.node_file)

        query = DrbXQuery("declare variable $x external; "
                          "declare variable $y external := 5; $x+$y")

        list_var = {'x': 9, 'y': 12}

        node_result = query.execute(self.node, list_var)

        self.assertEqual(len(node_result), 1)
        self.assertEqual(node_result[0], 21)

    def test_xquery_is(self):

        query = "declare variable $a := <a/>;" \
                "\ndeclare function local:testDoubleNodeIdentity(" \
                "$a as node(), " \
                "$b as node())" \
                "\n{\n$a is $b\n};" \
                "\nlocal:testDoubleNodeIdentity( <a/>, <a/>)," \
                "\nlocal:testDoubleNodeIdentity($a, $a)"
        drb_query = DrbXQuery(query)

        node_result = drb_query.execute(None)

        self.assertListEqual(node_result, [False, True])

    def test_xquery_from_node(self):
        xquery_file = self.current_path / "files" / "xquery_sample.xq"
        xquery_file = str(xquery_file)

        node_query = DrbFileFactory().create(xquery_file)

        drb_query = DrbXQuery(node_query)

        node_result = drb_query.execute(None)

        self.assertListEqual(node_result, [False, True])



from .drb_xquery_utils import DrbQueryFuncUtil
from .execptions import DynamicException, ErrorXQUERY


class XqueryBuildIn:
    def __init__(self, name,
                 exp_func,
                 namespace=None,
                 specific_context=None):
        self.name = name
        self.namespace = namespace
        self.exp_func = exp_func
        self.specific_context = specific_context


class XqueryVariable:

    def __init__(self, name, namespace=None, external=False,
                 value=None, type_var=None):
        self.namespace = namespace
        self.name = name
        self.external = external
        self.value = value
        if type_var is not None:
            self.type_var_namespace, self.type_var = \
                DrbQueryFuncUtil.split_namespace_name(type_var)
        else:
            self.type_var_namespace, self.type_var = (None, None)
        self._typed_value = None

    def get_value(self):
        if self._typed_value is None:
            self._typed_value = self.get_value_typed(self.type_var,
                                                     self.type_var_namespace,
                                                     self.value)
        return self._typed_value

    @staticmethod
    def get_value_typed(type_var, type_var_namespace, value,
                        arg_min: int = 1,
                        arg_max: int = 0,
                        func_msg: str = ''):
        if type_var is not None:
            if value is None:
                if arg_min == 0:
                    return ''
                else:
                    raise DynamicException(ErrorXQUERY.XPTY0004,
                                           func_msg +
                                           " empty value instead of" +
                                           type_var_namespace + ':' + type_var)
            from drb_xquery.drb_xquery_func import DrbQueryFuncCall
            if type_var_namespace == 'xs':
                if isinstance(value, list) and arg_max > 1:
                    result = []
                    for item in value:
                        result.append(DrbQueryFuncCall.xs_type_func[type_var](
                            type_var, None, None, item))
                else:
                    if isinstance(value, list):
                        if len(value) == 0:
                            if arg_min == 0:
                                return value
                            else:
                                raise DynamicException(ErrorXQUERY.XPTY0004,
                                                       func_msg +
                                                       " empty" +
                                                       " sequence " +
                                                       "but it is not " +
                                                       "allowed by definition")
                        if len(value) > 1:
                            raise DynamicException(ErrorXQUERY.XPTY0004,
                                                   func_msg +
                                                   " more than one " +
                                                   "but it is not " +
                                                   "allowed by definition")
                        value = value[0]
                    result = DrbQueryFuncCall.xs_type_func[type_var](
                        type_var, None, None, value)
                if result is not None or arg_min == 0:
                    return result
            raise DynamicException(ErrorXQUERY.XPST0051,
                                   func_msg +
                                   "Type " +
                                   type_var +
                                   "is unknown with namespace " +
                                   type_var_namespace)

        return value



from enum import Enum

from drb.exceptions import DrbException


class ErrorXQUERY(Enum):
    XPDY0002 = "Dynamic context is absent"
    XPST0017 = "No matching function call"
    XPTY0004 = "Type error"
    XPTY0020 = "Context item is not a node"
    FOAP0001 = "Wrong number of arguments."
    FOAR0001 = "Division by zero."
    FOCH0004 = "Collation does not support collation units."
    FORG0006 = "Invalid argument type."
    XPST0008 = "Unreferenced element, variable, attribute name or namespace"
    FODC0002 = "Error retrieving resource."
    FODC0005 = "Invalid argument to fn:doc or fn:doc-available."
    FOTY0012 = "Argument to fn:data() contains a node that does not have" \
               " a typed value."
    FORG0001 = "Invalid value for cast / constructor."
    FORX0001 = "Invalid regular expression flags."
    FORX0002 = "Invalid regular expression."
    FORX0003 = "Regular expression matches zero - length string."
    FORX0004 = "Invalid replacement string."
    XPST0051 = "Undefined type"
    XQST0134 = "The namespace axis is not supported."
    XPST0003 = "Syntax error."
    XQST0033 = "Duplicated namespace declaration"
    XQST0070 = "Undeclarable namespace"


class DynamicException(DrbException):
    def __init__(self, error: ErrorXQUERY, message):
        super().__init__(repr(error) + message)


class StaticException(DrbException):
    def __init__(self, error: ErrorXQUERY, message):
        super().__init__(repr(error) + message)



from .drb_xquery import DrbXQuery

from . import _version

__version__ = _version.get_versions()['version']

from .drb_xquery_context import DynamicContext

del _version

__all__ = [
    'DrbXQuery', 'DynamicContext'
]
